import logging
from typing import List, Optional, Dict, Type


from pycognaize.common.enums import (
    IqDocumentKeysEnum,
    ID,
    IqFieldKeyEnum,
    IqDataTypesEnum
)
from pycognaize.document.field import Field
from pycognaize.document.tag import ExtractionTag
from pycognaize.document.page import Page


class DateField(Field):
    """Base class for all pycognaize date fields"""
    tag_class: Type[ExtractionTag] = ExtractionTag

    def __init__(self,
                 name: str,
                 value: str = '',
                 tags: Optional[List[ExtractionTag]] = None,
                 field_id: Optional[str] = None,
                 group_key: str = None,
                 confidence: Optional[float] = -1.0
                 ):
        super().__init__(name=name, tags=tags, group_key=group_key,
                         confidence=confidence)
        self._field_id = field_id
        self._value = '; '.join([i.raw_value
                                 for i in self.tags]) if self.tags else value

    @property
    def field_id(self):
        return self._field_id

    @property
    def value(self):
        return self._value

    @classmethod
    def construct_from_raw(
            cls, raw: dict, pages: Dict[int, Page]) -> 'DateField':
        """Create DateField object from dictionary"""
        tag_dicts: List[dict] = raw[IqDocumentKeysEnum.tags.value]
        tags = []
        for i in tag_dicts:
            try:
                tags.append(cls.tag_class.construct_from_raw(
                    raw=i, page=pages[i['page']]))
            except Exception as e:
                logging.debug(f"Failed creating tag for field {raw[ID]}: {e}")
        return cls(name=raw[IqDocumentKeysEnum.name.value],
                   value=raw[IqFieldKeyEnum.value.value],
                   tags=tags,
                   field_id=str(raw[ID]),
                   group_key=raw.get(IqFieldKeyEnum.group_key.value, '')
                   )

    def to_dict(self) -> dict:
        """Converts DateField object to dictionary"""
        field_dict = super().to_dict()
        field_dict[IqFieldKeyEnum.value.value] = self.value
        field_dict[ID] = self._field_id
        field_dict[IqFieldKeyEnum.data_type.value] = IqDataTypesEnum.date.value
        return field_dict

    def __repr__(self):
        return (f"<{self.__class__.__name__}: {self.name}:"
                f" {'|'.join([i.raw_value for i in self.tags])}>")

    def __str__(self):
        return f"{'|'.join([i.raw_value for i in self.tags])}"



import warnings
import pycognaize


def module_not_found(stack_level: int = 2):
    """
    Warns user that function uses a module that is not listed in
    requirements.txt and is not installed in the environment

    :param stack_level: stack level of the warning
    :type stack_level: int
    """
    def module_not_found_custom(func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ModuleNotFoundError as e:
                message = f"Please install module " \
                          f"{str(e).split(' ')[-1]}. \n"\
                          f"It can be found in model-requirements.txt"
                warnings.warn(message, UserWarning, stacklevel=stack_level)
        return wrapper
    return module_not_found_custom


def soon_be_deprecated(version: str = None, stack_level: int = 2):
    """
    Warns user that the function they use will soon be deprecated.


    :param version: version after which the function will be deprecated
    :type version: str
    :param stack_level: stack level of the warning
    :type stack_level: int
    """
    if version is None:
        old_version = pycognaize.__version__.split(".")
        version = f"{old_version[0]}.{str(int(old_version[1]) + 1)}.0"

    def soon_be_deprecated_custom(func):
        def wrapper(*args, **kwargs):
            message = f"This function will be deprecated after " \
                      f"version {version}"
            warnings.warn(message, DeprecationWarning, stacklevel=stack_level)
            return func(*args, **kwargs)
        return wrapper
    return soon_be_deprecated_custom



import json
import logging
import os
import re
from typing import Optional, List, Iterable, Union
import numpy as np
from pycognaize.common.decorators import module_not_found


from pycognaize.common.enums import (
    StorageEnum,
    OCR_DATA_EXTENSION,
    IMG_EXTENSION
)
import pycognaize.common
from pycognaize.common.utils import (
    infer_rows_from_words,
    clean_ocr_data,
    find_first_word_coords,
    intersects,
    compute_intersection_area,
    stick_word_boxes,
    preview_img,
    image_string_to_array
)
from cloudstorageio import CloudInterface
from pycognaize.document.tag import ExtractionTag


class Page:
    """Representing a page of a document in pycognaize"""
    REGEX_NO_ALPHANUM_CHARS = re.compile(r'[^a-zA-Z\d)\[\](-.,]')

    def __init__(self, page_number: int, document_id: str, path: str):
        """

        :param page_number: The number of the page (1-based index)
        :param document_id: The unique id of the document
        :param path:
        """
        self._page_number = int(page_number)
        self._document_id = document_id
        self.ci = CloudInterface()
        self._path = path
        self._ocr_raw = None
        self._ocr = None
        self._lines = None
        self._row_word_groups = None
        self._image_bytes = None
        self._image_arr = None
        self._image_height = None
        self._image_width = None

    @property
    def page_number(self):
        """Page number of page"""
        return self._page_number

    @property
    def path(self):
        """Path of the source document"""
        return self._path

    @property
    def doc_id(self):
        """Document id of the page"""
        return self._document_id

    @property
    def ocr(self) -> dict:
        """Formatted ocr of page"""
        if self._ocr is None:
            self._ocr = self.get_ocr_formatted()
        return self._ocr

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.page_number}>"

    def get_image(self) -> bytes:
        """Converts image of page in bytes"""
        uri = os.path.join(self.path, StorageEnum.image_folder.value,
                           f"image_{self._page_number}.{IMG_EXTENSION}")
        try:
            with self.ci.open(uri, 'rb') as f:
                image_bytes = f.read()
        except FileNotFoundError as e:
            logging.debug(
                f"Unable to get the image for page {self.page_number}: {e}")
            with open(os.path.join(
                    os.path.dirname(pycognaize.common.__file__),
                    'white_pixel.jpeg'), 'rb') as f:
                image_bytes = f.read()
        return image_bytes

    @property
    def image_bytes(self) -> bytes:
        """Image of page in bytes"""
        if self._image_bytes is None:
            self._image_bytes = self.get_image()
        return self._image_bytes

    @property
    def image_arr(self) -> np.ndarray:
        """Numpy array of the page image"""
        if self._image_arr is None:
            self._image_arr = image_string_to_array(self.image_bytes)
        return self._image_arr

    @property
    def image_height(self) -> int:
        """Height of the page image"""
        if self._image_height is None:
            self._image_height = self.image_arr.shape[0]
        return self._image_height

    @property
    def image_width(self) -> int:
        """Width of the page image"""
        if self._image_width is None:
            self._image_width = self.image_arr.shape[1]
        return self._image_width

    @property
    def ocr_raw(self) -> dict:
        if self._ocr_raw is None:
            self._ocr_raw = self.get_ocr()
        return self._ocr_raw

    def get_ocr(self) -> Optional[dict]:
        """OCR of the page"""
        if self.path is None:
            raise ValueError("No path defined for getting the images")
        uri = os.path.join(self.path, StorageEnum.ocr_folder.value,
                           f"page_{self._page_number}.{OCR_DATA_EXTENSION}")
        try:
            with self.ci.open(uri, 'r') as f:
                ocr_raw = json.loads(f.read())
                ocr_raw['page']['height'] = float(ocr_raw['page']['height'])
                ocr_raw['page']['width'] = float(ocr_raw['page']['width'])
                for word in ocr_raw['data']:
                    word['x'] = float(word['x'])
                    word['y'] = float(word['y'])
                    word['w'] = float(word['w'])
                    word['h'] = float(word['h'])
        except FileNotFoundError as e:
            logging.debug(
                f"Unable to get the ocr for page {self.page_number}: {e}")
            ocr_raw = {
                "page": {"number": self.page_number,
                         "width": 1, "height": 1},
                "image": {"width": 1, "height": 1},
                "data": []}
        return ocr_raw

    def get_ocr_formatted(self, stick_coords: bool = False,
                          return_tags: bool = False
                          ) -> Union[dict, List[ExtractionTag]]:
        """Dict of words, paragraphs each containing their tag data"""
        res = dict(words=[], paragraphs=[])
        res_words_tags = []
        ocr = self.ocr_raw
        if ocr is None:
            return res
        image_height = float(self.image_height)
        image_width = float(self.image_width)
        page_height = float(ocr['page']['height'])
        page_width = float(ocr['page']['width'])
        if image_width > image_height and page_width < page_height:
            page_height, page_width = page_width, page_height
        height_ratio = image_height / page_height
        width_ratio = image_width / page_width
        for n, i in enumerate(ocr['data']):
            if not i['value'].strip():
                continue
            word = dict(
                left=round(float(i['x']) * width_ratio),
                right=round((float(i['x']) + float(i['w'])) * width_ratio),
                top=round(float(i['y']) * height_ratio),
                bottom=round((float(i['y']) + float(i['h'])) * height_ratio),
                ocr_text=i['value'],
                word_id_number=n)
            if word['left'] >= word['right']:
                word['right'] = word['left'] + 1
            if word['top'] >= word['bottom']:
                word['bottom'] = word['top'] + 1
            res['words'].append(word)
        if stick_coords:
            stick_word_boxes(box_coord=res['words'],
                             img_bytes=self.get_image())
        if return_tags:
            for word in res['words']:
                res_words_tags.append(self.word_to_extraction_tag(word))
            return res_words_tags
        return res

    @property
    def lines(self) -> List[List[dict]]:
        """Detects lines of the page

        :return: list of lists of dicts,
            where each list represents a line,
            and each dict in that list is a word on that line,
            with its coordinates, ocr_text and word_id_number
        """
        if self._lines is None:
            self._lines = self._create_lines()
        return self._lines

    def _create_lines(
            self,
            return_tags: bool = False
    ) -> List[List[Union[dict, ExtractionTag]]]:
        """Detects lines of the page

        :param return_tags: if False returns list of lists of dicts,
            where each list represents a line, and each dict
        in that list is a word on that line. If True,
            returns list of lists of ExtractionTags, where each word of line is
        converted to tag
        """
        words_tags = []
        rows_inf, self._row_word_groups = infer_rows_from_words(
            box=dict(left=0, top=0, right=1, bottom=1),
            class_ocr_data=clean_ocr_data(self.ocr)['words'])
        rows: list = [
            int(round(abs(rows_inf[row_n + 1]['top'] + row['bottom']) / 2))
            for row_n, row in enumerate(rows_inf[:-1])]
        bottom_coord = [int(rows_inf[-1]['bottom'])] if rows_inf else []
        self._row_word_groups: list = [
            i for _, i in sorted(
                zip(rows + bottom_coord, self._row_word_groups),
                key=lambda pair: pair[0])]
        rows.sort()
        temp_rows = []
        for i, (row, group) in enumerate(zip(rows + bottom_coord,
                                             self._row_word_groups)):
            if row in temp_rows:
                rows[i] = None
                orig_idx = rows.index(row)
                self._row_word_groups[orig_idx] += self._row_word_groups[i]
                self._row_word_groups[i] = None
            temp_rows.append(row)
        self._row_word_groups = [i for i in self._row_word_groups if i]
        self._row_word_groups = [
            sorted(i, key=lambda x: (x['left'], x['right']))
            for i in self._row_word_groups]
        if return_tags:
            for line in self._row_word_groups:
                new_line_tags = []
                for word in line:
                    new_line_tags.append(self.word_to_extraction_tag(word))
                words_tags.append(new_line_tags)
            return words_tags
        return self._row_word_groups

    def search_text(self, text: str,
                    case_sensitive=False,
                    sort: bool = False,
                    clean: bool = True,
                    area: dict = None,
                    cleanup_regex=REGEX_NO_ALPHANUM_CHARS,
                    return_tags: bool = False) -> list:
        """
        Detects the coordinates of the `text` in ocr of the page
            If the `text` is not found in the page return None

        :param str text:
        :param case_sensitive: If True, the search will be case-sensitive
        :param sort: If True,
            ocr_data will be ordered by `word_id_number` key before searching
        :param clean: If true,
            disregard all non-alphanumeric character from the search
        :param area: If a dict with coordinates (pixels) is given
            only search for text in specified area
        :param re._pattern_type cleanup_regex: Optional.
            Provide the regex for cleanup to be used
            (has effect only if `clean=True`)
        :param return_tags: if True, the words in found text
            are converted into tags.
        :return: List of dictionaries with word coordinates
            (keys: `left`, `right`, `top`, `bottom`, `matched_words`.
            `matched_words` includes the original word coordinate
                data for the matched words)
        :rtype: list
        """

        all_matches = []
        all_matches_tags = []
        if area:
            ocr_data = self.extract_area_words(**area)
        else:
            ocr_data = [word for line in self.lines for word in line]

        while True:
            final_coords = find_first_word_coords(
                text=text, ocr_data=ocr_data,
                case_sensitive=case_sensitive,
                sort=sort, clean=clean,
                cleanup_regex=cleanup_regex)
            if final_coords is None:
                break
            all_matches.append(final_coords)
            final_coord_word_ids = [
                i['word_id_number'] for i in final_coords['matched_words']]
            ocr_data = [word for word in ocr_data
                        if word['word_id_number'] not in final_coord_word_ids]
        if return_tags:
            for match in all_matches:
                matched_words_tags = []
                for word in match['matched_words']:
                    matched_words_tags.append(
                        self.word_to_extraction_tag(word))
                all_matches_tags.append(matched_words_tags)
            return all_matches_tags
        return all_matches

    @staticmethod
    def _validate_box_coordinates(left: [int, float],
                                  right: [int, float],
                                  top: [int, float],
                                  bottom: [int, float], ) -> None:
        """Validate coordinates of a box (with the origin on top-left corner)

        :param left: Left border of the box
        :param right: Right border of the box
        :param top: Top border of the box
        :param bottom: Bottom border of the box
        :return: Raises error if validation fails, otherwise does not return
            anything
        """

        if left < 0 or right < 0 or top < 0 or bottom < 0:
            raise ValueError('Coordinates need to be positive')
        if left > right:
            raise ValueError(
                f"Left ({left}) cannot be bigger than right ({right})")
        if top > bottom:
            raise ValueError(
                f"Top ({top}) cannot be bigger than bottom ({bottom})")
        if left == right:
            raise ValueError(
                f"Left ({left}) cannot be equal to right ({right})")
        if top == bottom:
            raise ValueError(
                f"Top ({top}) cannot be equal to bottom ({bottom})")

    def extract_area_words(self, left: [int, float],
                           right: [int, float],
                           top: [int, float],
                           bottom: [int, float],
                           threshold: float = 0.5,
                           return_tags: bool = False) -> Optional[list]:
        """Finds the words on the page which are included in the area
            resulted from given coordinates.

        :param threshold: Threshold value as a fraction
            (value between 0 and 1), default value is 0.5
        :param left: left coordinate
        :param right: right coordinate
        :param top: top coordinate
        :param bottom: bottom coordinate
        :param return_tags: if True, returns tags of the words
            embedded in given area
        :return: list of words, each element in the list is dictionary
            representing the coordinates, ocr_text of word, and word_id_number
        """
        if not 0 <= threshold <= 1:
            raise ValueError('Value of threshold must be between 0 and 1')
        self._validate_box_coordinates(
            left=left, right=right, top=top, bottom=bottom)
        word_list = []
        ocr_words = [word for line in self.lines for word in line]
        for word in ocr_words:
            if intersects(word,
                          left=left,
                          right=right,
                          top=top,
                          bottom=bottom):
                intersect_area = compute_intersection_area(word,
                                                           left=left,
                                                           right=right,
                                                           top=top,
                                                           bottom=bottom)
                word_area = (word['right'] - word['left']
                             ) * (word['bottom'] - word['top'])
                ratio = float(intersect_area / word_area)
                if ratio > threshold:
                    if return_tags:
                        word_list.append(self.word_to_extraction_tag(word))
                    else:
                        word_list.append(word)
        return word_list

    @staticmethod
    @module_not_found()
    def draw_rectangle(img: np.ndarray, left: int,
                       top: int, right: int, bottom: int) -> np.ndarray:
        """Adds a rectangle outline on the image"""
        import cv2
        return cv2.rectangle(img=img,
                             pt1=(int(left), int(top)),
                             pt2=(int(right), int(bottom)),
                             color=(0, 250, 0),
                             thickness=2)

    def draw_ocr_boxes(self, img: Optional[np.ndarray] = None) -> np.ndarray:
        """Draw boxes where text was detected and return the modified
            numpy array image

        :param img: Input image as numpy array.
            If not provided, use a copy of the instance image
        :return: numpy array of the image with word boxes
        """
        if img is None:
            img = self.image_arr.copy()
        for word in self.ocr['words']:
            img = self.draw_rectangle(img=img,
                                      left=word['left'],
                                      right=word['right'],
                                      top=word['top'],
                                      bottom=word['bottom'])

        return img

    @module_not_found()
    def draw_ocr_text(self, img: Optional[np.ndarray] = None) -> np.ndarray:
        """
        Insert corresponding OCR text above all OCR-ed words and
            return the modified numpy array image

        :param img: Input image as numpy array.
            If not provided, use a copy of the instance image
        :return: numpy array of the image with words
        """
        import cv2

        if img is None:
            img = self.image_arr.copy()
        for word in self.ocr['words']:
            img = cv2.putText(img=img, text=word['ocr_text'],
                              org=(word['left'], word['top'] - 1),
                              fontFace=cv2.FONT_HERSHEY_PLAIN,
                              fontScale=1.0, color=50, thickness=1)
        return img

    @module_not_found()
    def draw(self, fields: Optional[Iterable] = None,
             draw_ocr_boxes: bool = True,
             draw_ocr_text: bool = True,
             preview: bool = True,
             save: Optional[str] = None,
             draw_on_image: bool = False,
             size: int = 1000) -> np.ndarray:
        """
        Draw the fields and optionally draw the page image

        :param fields: List of Fields to draw
        :param draw_ocr_boxes: Whether to draw ocr boxes
        :param draw_ocr_text: Whether to draw ocr text
        :param preview: If True, show the image
        :param save: If True, the image will be saved to the given full path
        :param draw_on_image: If True, draw Tags on image instead of page
        :param size: Size of the preview image
        """
        import cv2
        if fields is None:
            fields = []
        img = self.image_arr.copy()
        if draw_ocr_boxes:
            img = self.draw_ocr_boxes(img=img)
        if draw_ocr_text:
            img = self.draw_ocr_text(img=img)

        draw_on_image = img if draw_on_image else None
        if size <= 0:
            raise ValueError(
                f"`size` argument must have a positive integer value,"
                f" got: {size}")
        for field in fields:
            for tag in field.tags:
                if self.page_number == tag.page_number:
                    img = tag.draw(draw_on_image)
        if preview:
            preview_img(img, size=size)
        if save:
            cv2.imwrite(filename=save, img=img)
        return img

    def free_form_text(self) -> str:
        """Return a text string from the ocr dictionary"""
        if self._row_word_groups is None:
            self._create_lines()
        final_text = '\n'.join([' '.join([word['ocr_text'] for word in row])
                                for row in self._row_word_groups])
        return final_text

    def word_to_extraction_tag(self, word: dict) -> ExtractionTag:
        """ Construct ExtractionTag object from word

        :type word: dict with left, right, top, bottom coordinates,
            ocr_text and word_id_number"""
        return ExtractionTag(left=word['left'] / self.image_width * 100,
                             right=word['right'] / self.image_width * 100,
                             top=word['top'] / self.image_height * 100,
                             bottom=word['bottom'] / self.image_height * 100,
                             page=self,
                             raw_value=word['ocr_text'],
                             raw_ocr_value=word['ocr_text'])


def create_dummy_page(page_n: int = 1, path: str = '/DUMMY/PATH'):
    """Used in test classes"""
    return Page(page_number=page_n, document_id='DUMMY_ID', path=path)



"""This module defines the Document object,
which includes the input and output fields for the model,
as well as the OCR data and page images of the document"""
import copy
import itertools
import logging
import os
from collections import OrderedDict
from typing import Dict, List, Tuple, Any, Optional

import fitz
from cloudstorageio import CloudInterface
from fitz.utils import getColor, getColorList

from pycognaize.common.enums import IqDocumentKeysEnum
from pycognaize.document.field import FieldMapping
from pycognaize.document.field.field import Field
from pycognaize.document.page import Page
from pycognaize.document.tag import TableTag
from pycognaize.document.tag.cell import Cell
from pycognaize.document.tag.tag import Tag


class Document:
    """Definition of input and output for a single document,
     depending on a given model"""

    def __init__(self,
                 input_fields: 'OrderedDict[str, List[Field]]',
                 output_fields: 'OrderedDict[str, List[Field]]',
                 pages: Dict[int, Page],
                 metadata: Dict[str, Any]):
        self._metadata = metadata
        self._pages: Dict[int, Page] = pages
        self._x: OrderedDict[str, List[Field]] = input_fields
        self._y: OrderedDict[str, List[Field]] = output_fields

    @property
    def x(self) -> 'OrderedDict[str, List[Field]]':
        """Returns a dictionary, where keys are input field names
        and values are list of Field objects"""
        return self._x

    @property
    def y(self) -> 'OrderedDict[str, List[Field]]':
        """Returns a dictionary, where keys are output field names
        and values are list of Field objects"""
        return self._y

    @property
    def metadata(self) -> Dict[str, Any]:
        """Returns document metadata"""
        return self._metadata

    @property
    def id(self) -> str:
        """Returns the pycognaize id of the document"""
        return self.metadata['document_id']

    @property
    def document_src(self):
        """Returns the source of the document"""
        return self.metadata['src']

    @property
    def pages(self) -> Dict[int, Page]:
        """Returns a dictionary, where each key is the page number
        and values are Page objects"""
        return self._pages

    @staticmethod
    def get_matching_table_cells_for_tag(
            tag: Tag,
            table_tags: List[TableTag],
            one_to_one: bool
    ) -> List[Tuple[Tag, TableTag, Cell, float]]:
        """Create a list which includes the original extraction tag,
           the corresponding table tag and Cell objects
           and the IOU of the intersection

        :param tag: The `tag` for which matching
            table and cells should be found
        :param table_tags: List of `table_tag`s
        :param one_to_one: If true,
            for each tag only one corresponding cell will be returned
        :return: List of tuples,
            which include the original extraction tag,
            the corresponding table tag and Cell objects
            and the IOU of the intersection
        """
        intersection = []
        if isinstance(tag, TableTag):
            return []
        for ttag in table_tags:
            if ttag.page.page_number != tag.page.page_number:
                continue
            for cell in ttag.cells.values():
                temp_cell = copy.deepcopy(cell)
                temp_cell.page = ttag.page
                iou = tag.iou(temp_cell)
                if iou <= 0:
                    continue
                if one_to_one:
                    if (not intersection or (
                            intersection and intersection[0][-1] < iou
                    )):
                        intersection = [(tag, ttag, cell, iou)]
                else:
                    intersection.append((tag, ttag, cell, iou))
        return intersection

    def get_table_cell_overlap(
            self, source_field: str,
            one_to_one: bool) -> List[Tuple[Tag, TableTag, Cell, float]]:
        """Create a list which includes the original extraction tag,
           the corresponding table tag and Cell objects
           and the IOU of the intersection

        :param source_field: Name of the field,
            for which to return the corresponding table cells
        :param one_to_one: If true,
            for each tag only one corresponding cell will be returned
        :return: List of tuples,
            which include the original extraction tag,
            the corresponding table tag and Cell objects
            and the IOU of the intersection
        """
        # noinspection PyUnresolvedReferences
        table_tags = [
            tag
            for fields in itertools.chain(self.x.values(), self.y.values())
            for field in fields
            for tag in field.tags
            if isinstance(tag, TableTag)
        ]
        res = []
        if source_field in self.x:
            fields = self.x[source_field]
        elif source_field in self.y:
            fields = self.y[source_field]
        else:
            return []
        for field in fields:
            for tag in field.tags:
                intersection = self.get_matching_table_cells_for_tag(
                    tag=tag, table_tags=table_tags, one_to_one=one_to_one)
                res.extend(intersection)
        return res

    def to_dict(self) -> dict:
        """Converts Document object to dict"""
        input_fields = OrderedDict(
            {name: [field.to_dict() for field in fields]
             for name, fields in self.x.items()})
        output_fields = OrderedDict(
            {name: [field.to_dict() for field in fields]
             for name, fields in self.y.items()})
        data = OrderedDict(input_fields=input_fields,
                           output_fields=output_fields,
                           metadata=self.metadata)
        return data

    @classmethod
    def from_dict(cls, raw: dict, data_path: str) -> 'Document':
        """Document object created from data of dict
        :param raw: document dictionary
        :param data_path: path to the documents OCR and page images
        """
        if not isinstance(raw, dict):
            raise TypeError(
                f"Expected dict for 'raw' argument got {type(raw)} instead")
        metadata = raw['metadata']
        pages = OrderedDict({page_n: Page(page_number=page_n,
                                          document_id=metadata['document_id'],
                                          path=data_path)
                             for page_n in range(
                1, metadata['numberOfPages'] + 1)})
        input_fields = OrderedDict(
            {name: [
                FieldMapping[
                    field[IqDocumentKeysEnum.data_type.value]
                ].value.construct_from_raw(raw=field, pages=pages)
                for field in fields]
             for name, fields in raw['input_fields'].items()})
        output_fields = OrderedDict(
            {name: [
                FieldMapping[
                    field[IqDocumentKeysEnum.data_type.value]
                ].value.construct_from_raw(raw=field, pages=pages)
                for field in fields]
             for name, fields in raw['output_fields'].items()})
        return cls(input_fields=input_fields, output_fields=output_fields,
                   pages=pages, metadata=metadata)

    def _collect_all_tags_for_fields(self,
                                     field_names: List[str],
                                     is_input_field: bool = True) -> List[Tag]:
        """Collect all tags of given field names from either input or output
            fields

        :param field_names: List of strings representing the field names
        :param is_input_field: If true, collect tags from input fields,
            otherwise collect tags from output fields
        :return: List of tags from the specified fields
        """
        all_tags = []
        if is_input_field:
            field_dict = self.x
            field_type = 'input field'
        else:
            field_dict = self.y
            field_type = 'output field'
        if field_names is not None:
            for field_name in field_names:
                if field_name not in field_dict.keys():
                    raise ValueError(f'Invalid {field_type} {field_name}')
                for field in field_dict.get(field_name, []):
                    for tag in field.tags:
                        all_tags.append(tag)
        return all_tags

    def to_pdf(self,
               input_fields: Optional[List[str]] = None,
               output_fields: Optional[List[str]] = None,
               input_color: str = 'deeppink1',
               output_color: str = 'deepskyblue3',
               input_opacity: float = 0.2,
               output_opacity: float = 0.3) -> bytes:

        """
        Adds tags of input_fields and output_fields to the bytes object
        representing the pdf file of the document.

        :param input_fields: Input fields
        :param output_fields: Output fields
        :param input_color: The color of the annotation rectangle
            of the input field
        :param output_color: The color of the annotation rectangle
            of the output field
        :param input_opacity: The opacity of the annotation rectangle
            of the input field
        :param output_opacity: The opacity of the annotation rectangle
            of the output field
        :return: bytes object of the pdf
        """

        ci = CloudInterface()
        pdf_path = os.path.join(self.pages[1].path, self.document_src) + '.pdf'

        with ci.open(pdf_path, 'rb') as f:
            pdf_bytes = f.read()
        doc_fitz = fitz.open('pdf', pdf_bytes)

        if input_fields is not None:
            input_tags = self._collect_all_tags_for_fields(
                field_names=input_fields, is_input_field=True)
            for tag in input_tags:
                pdf_bytes = annotate_pdf(
                    doc=doc_fitz,
                    tag=tag,
                    color=input_color,
                    opacity=input_opacity)
        if output_fields is not None:
            input_tags = self._collect_all_tags_for_fields(
                field_names=output_fields, is_input_field=False)
            for tag in input_tags:
                pdf_bytes = annotate_pdf(
                    doc=doc_fitz,
                    tag=tag,
                    color=output_color,
                    opacity=output_opacity)
        return pdf_bytes


def annotate_pdf(doc: fitz.Document,
                 tag: Tag,
                 color: str,
                 opacity: float = 0.3) -> bytes:
    """An annotated Document pdf in bytes"""
    page = doc[tag.page.page_number-1]
    x0 = tag.left * page.MediaBox.width / 100
    y0 = tag.top * page.MediaBox.height / 100
    x1 = tag.right * page.MediaBox.width / 100
    y1 = tag.bottom * page.MediaBox.height / 100
    annot_rect = fitz.Rect(x0, y0, x1, y1)
    if color.upper() not in getColorList():
        raise ValueError(f'Wrong color {color}')
    if opacity < 0 or opacity > 1:
        raise ValueError(f'Wrong opacity value {opacity}')
    color_dict = {"stroke": getColor(color), "fill": getColor(color)}
    annot = page.add_rect_annot(annot_rect)
    annot.set_colors(color_dict)
    annot.set_opacity(opacity)
    annot.update()
    return doc.write()


def filter_out_invalid_tables(tables):
    valid_tables = []
    for table in tables:
        if not table.tags:
            logging.warning('removing table with no tags')
            continue

        valid_tables.append(table)
    return valid_tables


def assign_indices_to_tables(tables: List['Field']):
    tables_dict = {}
    valid_tables = filter_out_invalid_tables(tables)
    tables = sorted(valid_tables, key=lambda x: (
        x.tags[0].page.page_number, x.tags[0].top, x.tags[0].left))

    for table in tables:
        page_number = table.tags[0].page.page_number
        try:
            # find all tables in the current page
            # and get the index for next table
            table_last_index = [
                current_index
                for (table_page_number, current_index) in tables_dict.keys()
                if page_number == table_page_number
            ][-1] + 1
        except IndexError:
            table_last_index = 0
        tables_dict[(page_number, table_last_index)] = table
    return tables_dict



import abc

from typing import List, Optional, Dict, Type

from pycognaize.common.enums import IqFieldKeyEnum
from pycognaize.document.page import Page
from pycognaize.document.tag.tag import Tag


class Field(metaclass=abc.ABCMeta):
    """Base class for all Field types"""
    tag_class: Type[Tag] = Tag

    def __init__(self, name: str, tags: Optional[List[Tag]] = None,
                 group_key: Optional[str] = None,
                 confidence: Optional[float] = -1.0
                 ):
        self._confidence = confidence
        if group_key is None:
            group_key = ''
        self._group_key = group_key
        self._name = name
        if tags is None:
            self._tags = []
        else:
            self._tags = tags

    @property
    def name(self):
        return self._name

    @property
    def tags(self):
        return self._tags

    @property
    def group_key(self):
        return self._group_key

    @group_key.setter
    def group_key(self, value):
        if not isinstance(value, str):
            raise TypeError(f"expected str, got {type(value)}")
        self._group_key = value

    @property
    def confidence(self):
        return self._confidence

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name}>"

    def __str__(self):
        return f"{self.tags}"

    @classmethod
    @abc.abstractmethod
    def construct_from_raw(cls, raw: dict, pages: Dict[int, Page]) -> 'Field':
        """Use raw dictionary in order to recreate the Field python object"""
        pass

    @abc.abstractmethod
    def to_dict(self) -> dict:
        """Return a dictionary representing the field object"""
        field_dict = dict()
        field_dict[IqFieldKeyEnum.name.value] = self.name
        if self._group_key:
            field_dict[IqFieldKeyEnum.group_key.value] = self._group_key
        field_dict[IqFieldKeyEnum.tags.value] = [
            i.to_dict() for i in self.tags]
        return field_dict



import io
import os
import re

from PIL import Image
import bson
import logging
from itertools import groupby
from typing import Union, List, Optional
from bson.json_util import loads as bson_loads

import numpy as np
from dataclasses import dataclass
from pycognaize.common.enums import PythonShellEnum
from pycognaize.common.decorators import soon_be_deprecated


REGEX_NO_ALPHANUM_CHARS = re.compile(r'[^a-zA-Z\d)\[\](-.,]')


def is_float(str_number: str) -> bool:
    """Check if the string value is a valid number.
        Infinity and NaN are regarded as invalid"""
    # 'n' check covers both 'NaN', 'Inf' and 'Infinity'
    if 'n' in str_number.lower():
        return False
    try:
        float(str_number)
        return True
    except ValueError:
        return False


def convert_coord_to_num(val: Union[float, int, str]) -> Union[int, float]:
    """If the input is a string representation of a number
        (with or without a %), convert it into float.
        If the input is already a float or an integer, return itself"""
    if isinstance(val, str):
        val = val.rstrip('%')
        if not is_float(val):
            raise ValueError(f"Invalid numeric {val}")
        val = float(val)
    elif isinstance(val, (float, int)):
        pass
    else:
        raise TypeError(f"Invalid type for numeric {type(val)} ({val}")
    return val


def load_bson_by_path(doc_path):
    with open(os.path.join(doc_path), 'r') as f:
        return bson_loads(f.read())


@soon_be_deprecated()
def bytes_to_array(img_str: bytes) -> np.ndarray:
    """Convert image bytes into numpy array

    :param img_str: Bytestream of the image
    :return: numpy array of the image
    """
    import cv2
    nparr = np.frombuffer(img_str, np.uint8)
    img_np = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    return img_np


@soon_be_deprecated()
def string_to_array(img_str: bytes):
    """
    Convert a bytestring into a numpy array with opencv

    :param str img_str: image as byte string
    :return: img_np: image as numpy array
    """
    import cv2
    nparr = np.frombuffer(img_str, np.uint8)
    img_np = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    assert type(img_np) == np.ndarray, (
        "Expected numpy array, received {}".format(type(img_np)))

    return img_np


def image_bytes_to_array(img_str: bytes) -> np.ndarray:
    """Convert image bytes into numpy array

    :param img_str: Bytestream of the image
    :return: numpy array of the image
    """
    image = Image.open(io.BytesIO(img_str))
    # Remove <alpha> channel if it exists
    img_np = np.array(image)[..., :3]

    return img_np


def image_string_to_array(img_str: bytes):
    """
    Convert a bytestring into a numpy array with opencv

    :param str img_str: image as byte string
    :return: img_np: image as numpy array
    """
    image = Image.open(io.BytesIO(img_str))
    # Remove <alpha> channel if it exists
    img_np = np.array(image)[..., :3]
    assert type(img_np) == np.ndarray, (
        "Expected numpy array, received {}".format(type(img_np)))

    return img_np


def compute_otsu_threshold(img_array):
    bins_num = 256
    hist, bin_edges = np.histogram(img_array, bins=bins_num)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2.

    weight1 = np.cumsum(hist)
    weight2 = np.cumsum(hist[::-1])[::-1]

    mean1 = np.cumsum(hist * bin_centers) / weight1
    mean2 = (np.cumsum((hist * bin_centers)[::-1]) / weight2[::-1])[::-1]

    weight_mul = weight1[:-1] * weight2[1:]
    inter_class_variance = weight_mul * (mean1[:-1] - mean2[1:]) ** 2
    index_of_max_val = np.argmax(inter_class_variance)

    threshold = bin_centers[:-1][index_of_max_val]

    return threshold


def img_to_black_and_white(img_array):
    """Image to only black and white image

    black pixel --> white pixel
    other pixel --> black pixel

    :return:
    """

    def grayscale2bw(x, threshold_value):
        return 255 if x < threshold_value else 0

    vectorized_converter = np.vectorize(grayscale2bw)
    grayscale = np.dot(img_array[..., :3], [0.299, 0.587, 0.114])
    threshold = compute_otsu_threshold(img_array)
    return vectorized_converter(grayscale, threshold)


def group_sequence(list_of_integers):
    """Group sequence into list
    example ->
    test_list = [1, 2, 3, 8, 15, 23, 24, 25, 10, 11, 13, 15]
    sequence_list = [[1, 2, 3], [23, 24, 25], [10, 11]]
    :param list_of_integers: list of integers
    :return: list of grouped integers
    """
    it = iter(list_of_integers)
    prev, res = next(it), []

    while prev is not None:
        start = next(it, None)
        if prev + 1 == start:
            res.append(prev)
        elif res:
            yield list(res + [prev])
            res = []
        prev = start


def stick_word_boxes(box_coord: List[dict], img_bytes: bytes, padding=1):
    """Stick boxes with OpenCV
        1) Stick by y (height), add "2 pixel" padding in bottom and right
            Get bigger sequence of 255 < vectors_y < 0  --> bigger_sequence
            new_top = min(bigger_sequence)
            new_bottom = max(bigger_sequence)

        2) stick by x (width), add "2 pixel" padding in bottom
            Get numpy array of 255 < vectors_x < 0  --> min_max_x
            new_left = min(min_max_x)
            new_right = max(min_max_x)

    :param box_coord: List of dictionaries representing
        boxes with the following format:
        ({'left': <int>, 'right': <int>, 'top': <int>, 'bottom': <int>})
    :param img_bytes: The original image to which the coordinates correspond to
    :param padding: Expand the final boxes with this amount of pixels
        on each four sides
    :return:
    """
    np_image = image_bytes_to_array(img_bytes)
    b_and_w_image = img_to_black_and_white(np_image)
    h, w, _ = np_image.shape
    for word in box_coord:
        start_point_xmin = max(0, int(round(word['left'])) - 1)
        start_point_ymin = max(0, int(round(word['top'])) - 1)

        end_point_xmax = int(round(word['right'])) + 2
        end_point_ymax = int(round(word['bottom'])) + 2

        cropped_word = b_and_w_image[start_point_ymin:end_point_ymax,
                                     start_point_xmin:end_point_xmax]

        vectors_y = np.mean(cropped_word, axis=1)
        min_max_y = np.where((vectors_y != 0) & (vectors_y != 255))

        if len(min_max_y[0]) == 0:
            continue
        elif 0 < len(min_max_y[0]) < 3:
            vectors_x = np.mean(cropped_word, axis=0)
            min_max_x = np.where((vectors_x != 0) & (vectors_x != 255))

            new_top = min(min_max_y[0]) - 2 if len(min_max_y[0]) else 0
            new_bottom = max(min_max_y[0]) + 2 if len(min_max_y[0]) else 0
            new_left = min(min_max_x[0]) - 2 if len(min_max_x[0]) else 0
            new_right = max(min_max_x[0]) + 2 if len(min_max_x[0]) else 0
        else:
            try:
                sequence_y = list(group_sequence(list(min_max_y[0])))
                if len(sequence_y) > 0:
                    sequence_y = sorted(sequence_y, key=lambda x: len(x),
                                        reverse=True)
                max_sequence = sequence_y[0]
                new_top = min(max_sequence)
                new_bottom = max(max_sequence)
                new_bottom_cropped = new_bottom + 2

                cropped_word_height = cropped_word[
                                      new_top:new_bottom_cropped, :]

                vectors_x_height_stuck = np.mean(cropped_word_height, axis=0)
                min_max_x_height_stuck = np.where(
                    (vectors_x_height_stuck != 0
                     ) & (vectors_x_height_stuck != 255)
                )

                new_left = int(min(min_max_x_height_stuck[0]))
                new_right = int(max(min_max_x_height_stuck[0]))
            except Exception as e:
                logging.debug(f"Failed for word {word}: {e}")
                continue

        word['left'] = int(new_left) + start_point_xmin - padding
        word['right'] = int(new_right) + start_point_xmin + padding
        word['top'] = int(new_top) + start_point_ymin - padding
        word['bottom'] = int(new_bottom) + start_point_ymin + padding

    return box_coord


def preview_img(img: np.ndarray, size: int = 1000):
    """Preview the given image in a window

    :param img: Original image as numpy array
    :param size: Resize the window to the given resolution before viewing
    """
    import matplotlib.pyplot as plt
    plt.imshow(img)
    fig = plt.figure(num=1)
    fig.set_dpi(size)
    fig.set_size_inches(1, 1, forward=True)
    plt.axis('off')
    plt.show()


def infer_rows_from_words(box, class_ocr_data, auto_thresh=True, thresh=12,
                          min_height=10, min_width=10):
    """
    Infer row coordinates by class crop (table, column, row)
        from the ocr output
    :param class_ocr_data:
    :param box:
    :param auto_thresh:
    :param thresh:
    :param min_height:
    :param min_width:
    :rtype: tuple(list, list)
    :return:
    """
    class_ocr_data = sorted(class_ocr_data,
                            key=lambda x: (x['top'], x['left']))
    class_ocr_data = [
        i for i in class_ocr_data
        if i['ocr_text'].strip(' \t\n.,:;*_|\\/!\'"')]
    if auto_thresh:
        cleaned_ocr_data_for_height_detection = [
            i for i in class_ocr_data if
            i['ocr_text'].strip(' \t\n.,:;*_-~=|\\/!\'"')]
        all_heights_tuples = [
            (i.get('bottom') - i.get('top'), i['ocr_text']) for i in
            cleaned_ocr_data_for_height_detection
            if i.get('bottom') and i.get('top')]
        all_heights = [i[0] for i in all_heights_tuples]
        if len(all_heights) > 0:
            average_heights = sum(all_heights) / len(all_heights)
        else:
            average_heights = 0
        thresh = int(round(average_heights * 0.8))
        min_height = int(round(average_heights * 0.7))
        min_width = int(round(average_heights * 0.7))
    previous_bottom = 0
    previous_top = 0
    word_groups = []
    # Group aligned rows
    for i, (row_top, row_boxes) in enumerate(
            groupby(class_ocr_data, key=lambda x: x.get('top'))):
        word_boxes = sorted(list(row_boxes), key=lambda x: (x.get('left')))
        row_bottom = max(i['bottom'] for i in word_boxes)
        if not i:
            word_groups.append(word_boxes)
        elif (
                # (row_top - previous_bottom < thresh) or
                (abs(row_top - previous_top)
                 < thresh and abs(row_bottom - previous_bottom) < thresh) or
                (row_top >= previous_top and row_bottom <= previous_bottom) or
                (row_top <= previous_top and row_bottom >= previous_bottom)
        ):
            word_groups[-1] += word_boxes
        else:
            word_groups.append(word_boxes)
        previous_bottom = row_bottom
        previous_top = row_top
    aggregated_rows = []
    # Construct row boxes from row groups
    for row in word_groups:
        temp_d = dict(left=min(i['left'] for i in row) - box['left'],
                      right=max(i['right'] for i in row) - box['left'],
                      top=min(i['top'] for i in row) - box['top'],
                      bottom=max(i['bottom'] for i in row) - box['top'],
                      ocr_text=' '.join(i['ocr_text'] for i in row)
                      )
        if (temp_d['bottom'] - temp_d['top'] > min_width
                and temp_d['right'] - temp_d['left'] > min_height):
            aggregated_rows.append(temp_d)
    return aggregated_rows, word_groups


def clean_ocr_data(ocr_data: dict, thresh: float = 4.0) -> dict:
    """Cleans the ocr data

    :param ocr_data: OCR data dictionary, where values under the key `words`
        have the following form: {'top'}
    :param thresh: Words that
    :return:
    """
    ocr_data['words'] = [i for i in ocr_data['words'] if i['ocr_text'].strip()
                         and (i['bottom'] - i['top'])
                         / max(i['right'] - i['left'], 0.01) <= thresh]
    return ocr_data


def find_first_word_coords(text: str, ocr_data: list,
                           case_sensitive: bool = False, sort: bool = False,
                           clean: bool = True,
                           cleanup_regex=REGEX_NO_ALPHANUM_CHARS
                           ) -> Optional[dict]:
    """
    Detect the coordinates of the first occurrence
        of `text` in `ocr_data` if any.
    If the `text` is not found in `ocr_data` return None
    :param text:
    :param ocr_data: List of dictionaries.
    Each dictionary contains information about a single word.
    Each word dictionary has the following keys: `confidence`, `right`,
        `left`, `top`, `bottom`, `ocr_text`, `word_id_number`
    :param case_sensitive: If True, the search will be case-sensitive
    :param sort: If True, ocr_data will be ordered by `word_id_number`
        key before searching
    :param clean: If true, disregard all non-alphanumeric character
        from the search
    :param re._pattern_type cleanup_regex: Optional.
        Provide the regex for cleanup to be used
        (has effect only if `clean=True`)
    :return: Dictionary with word coordinates
        (keys: `left`, `right`, `top`, `bottom`, `matched_words`.
        `matched_words` includes the original word coordinate data
        for the matched words)
    """
    matched_words = []
    if sort:
        ocr_data = sorted(ocr_data, key=lambda x: float(x['word_id_number']))
    if not case_sensitive:
        text = text.lower()
    words = text.split(' ')
    idx = 0
    for ocr_d in ocr_data:
        ocr_word = ocr_d['ocr_text']
        if clean:
            ocr_word = cleanup_regex.sub('', ocr_word)
            word = cleanup_regex.sub('', words[idx])
        else:
            word = words[idx]
        ocr_word = ocr_word.strip()
        if not case_sensitive:
            ocr_word = ocr_word.lower()
        if ocr_word == word:
            idx += 1
            matched_words.append(ocr_d)
            if idx >= len(words):
                break
        else:
            matched_words = []
            idx = 0
    else:
        return
    return dict(top=min(matched_words, key=lambda x: x['top'])['top'],
                bottom=max(matched_words, key=lambda x: x['bottom'])['bottom'],
                left=min(matched_words, key=lambda x: x['left'])['left'],
                right=max(matched_words, key=lambda x: x['right'])['right'],
                matched_words=matched_words)


def intersects(word: dict,
               left: [int, float],
               right: [int, float],
               top: [int, float],
               bottom: [int, float]) -> bool:
    """
    Given the word and coordinates of the area,
        returns true if word intersects area, otherwise false
    :param word: dictionary, containing coordinates,
        ocr_text and word_id_number of the word
    :param left: left coordinate
    :param right: right coordinate
    :param top: top coordinate
    :param bottom: bottom coordinate
    :return: bool
    """
    return (word['left'] < right and word['right'] > left
            and word['top'] < bottom and word['bottom'] > top)


def compute_intersection_area(word: dict,
                              left: [int, float],
                              right: [int, float],
                              top: [int, float],
                              bottom: [int, float]) -> [int, float]:
    """
    Given the word and coordinates of the area, computes area of intersection
    :param word: dictionary, containing coordinates,
        ocr_text and word_id_number of the word
    :param left: left coordinate
    :param right: right coordinate
    :param top: top coordinate
    :param bottom: bottom coordinate
    :return: area of intersection
    """
    left = max([left, word['left']])
    top = max([top, word['top']])
    right = min([right, word['right']])
    bottom = min([bottom, word['bottom']])
    return (right - left) * (bottom - top)


def detect_python_shell() -> str:
    """Detect what python shell is being used - Standard Interpreter,
        Jupyter Notebook, Interactive shell or other"""
    try:
        # noinspection PyUnresolvedReferences
        shell = get_ipython().__class__.__name__
        if shell == PythonShellEnum.zmq_interactive_shell.value:
            # Jupyter notebook or qtconsole
            return PythonShellEnum.zmq_interactive_shell.value
        elif shell == PythonShellEnum.terminal_interactive_shell.value:
            # Terminal running IPython
            return PythonShellEnum.terminal_interactive_shell.value
        else:
            # Other type (?)
            return PythonShellEnum.other_type_shell.value
    except NameError:
        # Probably standard Python interpreter
        return PythonShellEnum.standard_python_interpreter.value


def replace_object_ids_with_string(bson_obj):
    if isinstance(bson_obj, dict):
        for key, value in bson_obj.items():
            bson_obj[key] = replace_object_ids_with_string(value)
    if isinstance(bson_obj, list):
        for idx, el in enumerate(bson_obj):
            bson_obj[idx] = replace_object_ids_with_string(el)
    if isinstance(bson_obj, bson.ObjectId):
        return str(bson_obj)
    else:
        return bson_obj


@dataclass
class ConfusionMatrix:
    TP: int = 0
    TN: int = 0
    FP: int = 0
    FN: int = 0

    @property
    def f1(self) -> float:
        try:
            return self.TP / (self.TP + 0.5 * (self.FP + self.FN))
        except ZeroDivisionError:
            return 0

    @property
    def precision(self) -> float:
        try:
            return self.TP / (self.TP + self.FP)
        except ZeroDivisionError:
            return 0

    @property
    def recall(self) -> float:
        try:
            return self.TP / (self.TP + self.FN)
        except ZeroDivisionError:
            return 0

    @property
    def accuracy(self) -> float:
        try:
            return (self.TP + self.TN) / (
                    self.FP + self.FN + self.TP + self.TN)
        except ZeroDivisionError:
            return 0

    def to_dict(self) -> dict:
        return {
            'TP': self.TP,
            'TN': self.TN,
            'FP': self.FP,
            'FN': self.FN
        }

    def compute_metrics(self) -> dict:
        """ Given confusion matrix, computes f1 score, precision,
            recall and accuracy.
         :return: results in dict"""
        result = {
            'confusion matrix': self.to_dict(),
            'f1 score': self.f1,
            'precision': self.precision,
            'recall': self.recall,
            'accuracy': self.accuracy}
        return result

    def draw_cm(self, title: str = 'Confusion Matrix'):
        cf_matrix = [self.TP, self.FP, self.FN, self.TN]
        cf_matrix = np.asarray(cf_matrix).reshape(2, 2)
        group_names = ['True Positive', 'False Positive',
                       'False Negative', 'True Negative']
        group_counts = ["{0:0.0f}".format(value)
                        for value in cf_matrix.flatten()]
        group_percentages = [
            "{0:.2%}".format(value)
            for value in cf_matrix.flatten() / np.sum(cf_matrix)]
        labels = [
            f'{v1}\n{v2}\n{v3}'
            for v1, v2, v3 in zip(group_names,
                                  group_counts,
                                  group_percentages)]
        labels = np.asarray(labels).reshape(2, 2)
        import seaborn as sns
        import matplotlib.pyplot as plt

        plt.figure(figsize=(12, 7))
        plt.title(title, fontsize=20)
        sns.heatmap(
            cf_matrix,
            annot=labels,
            fmt='',
            xticklabels=[1, 0],
            yticklabels=[1, 0],
            cmap='Blues',
            annot_kws={"size": 17})
        plt.show()


def get_index_of_first_non_empty_list(list_of_lists):
    non_empty_idx = len(list_of_lists)
    for idx, line in reversed(list(enumerate(list_of_lists))):
        if line:
            return idx
    return non_empty_idx


def convert_tag_coords_to_percentages(tag, w, h) -> dict:
    return dict(left=tag.left * w / 100,
                right=tag.right * w / 100,
                top=tag.top * h / 100,
                bottom=tag.bottom * h / 100)



import abc
from typing import Any
import requests

from pycognaize.document import Document


class Index(metaclass=abc.ABCMeta):
    """
    Use this abstract class for creating task specific Index subclasses,
    which allow comparison and matching between pycognaize documents
    """

    def __init__(self, token: str, url: str):
        self.url = url
        self.session = requests.Session()
        self.session.headers = {'x-auth': token}

    @property
    def id(self) -> str:
        return self.url.strip('/').rsplit('/', 1)[-1]

    def build_and_store(self, document: Document) -> requests.Response:
        """Builds an encoded document and stores it in the session"""
        doc_encoded = self.build(document)
        doc_idx = document.id
        return self._store(doc_id=doc_idx, encoding=doc_encoded)

    def _store(self, doc_id: str, encoding) -> requests.Response:
        return self.session.put(url=self.url,
                                json={'data': {doc_id: encoding}},
                                verify=False)

    @staticmethod
    def response_to_dict(response):
        """Returns a dictionary representing the response
        where keys are document ids and values are the encoded documents"""
        full_index = {}
        for i in response['findexes']:
            full_index[i['documentId']] = i['data']
        return full_index

    def match_and_get(self, document: Document) -> Document:
        """Match a given document with an existing document in the index"""
        get_response: requests.Response = self.session.get(self.url,
                                                           verify=False)
        full_index: dict = self.response_to_dict(get_response.json())
        if document.id not in full_index:
            encoding = self.build(document=document)
            self._store(doc_id=document.id, encoding=encoding)
            full_index[document.id] = encoding
        document = self.match(document=document, full_index=full_index)
        return document

    @abc.abstractmethod
    def build(self, document: Document) -> Any:
        """Build an encoding for a document, which can then be used for
        comparison and matching
        :param document: The document object to be built
        :return: The final encoding. The type of the encoding is task specific,
        therefore allows any type
        """
        raise NotImplementedError

    @abc.abstractmethod
    def match(self, document: Document, full_index: dict) -> Document:
        """Match a given document with an existing document in the index
        :param document: The document to be matched with an existing one
        :param full_index: The document encoding
        :return int: A tuple with the matched document, and a confidence value
        from 0 to 100
        """
        raise NotImplementedError



"""Defines LazyDocumentDict, used for lazy loading the documents
in the snapshot
"""

import logging
from typing import Optional

from bson import json_util
import os
from collections.abc import Mapping

from cloudstorageio import CloudInterface

from pycognaize.common.enums import StorageEnum
from pycognaize.document import Document


class LazyDocumentDict(Mapping):
    """Contains documents included in the snapshot"""
    CI = CloudInterface()
    document_filename = StorageEnum.doc_file.value

    def __init__(self, doc_path: str, data_path: str):
        self._doc_path = doc_path
        self._data_path = data_path
        self._ids = sorted([os.path.basename(os.path.dirname(i))
                           for i in self.CI.listdir(doc_path)
                           if self.CI.isfile(os.path.join(doc_path,
                            i, self.document_filename))])

    @property
    def doc_path(self) -> str:
        """Path of the document's JSON"""
        return self._doc_path

    @property
    def data_path(self) -> str:
        """Path of the document's OCR and page images"""
        return self._data_path

    def __getitem__(self, doc_id) -> Optional[Document]:
        """The Document object, retrieved from provided path

        Note: Path can be both local and remote
        """
        try:
            path = os.path.join(self.doc_path, doc_id,
                                f"{self.document_filename}")
            if self.CI.is_local_path(path):
                with open(path, 'r', encoding='utf8') as f:
                    doc_dict = json_util.loads(f.read())
            else:
                with self.CI.open(path, 'r') as f:
                    doc_dict = json_util.loads(f.read())
            return Document.from_dict(raw=doc_dict,
                                      data_path=os.path.join(self.data_path,
                                                             doc_id))
        except Exception as e:
            logging.error(f'Failed reading document {doc_id}: {e}')

    def __iter__(self):
        return iter(self._ids)

    def __len__(self):
        return len(self._ids)



import abc
from collections import Counter, defaultdict
from typing import Tuple, List

import requests
import simplejson as json
from cloudstorageio import CloudInterface

from pycognaize.common.utils import (
    replace_object_ids_with_string,
    ConfusionMatrix
)
from pycognaize.document.document import Document
from pycognaize.document.tag import ExtractionTag


def join_url(*parts):
    """Join parts into a single url string"""
    return '/'.join(s.strip('/') for s in parts)


class Model(metaclass=abc.ABCMeta):
    """Inherit from this abstract class and implement `predict` method

    The model inputs and outputs are available from the document attribute.
    """
    CI = CloudInterface()

    @abc.abstractmethod
    def predict(self, document: Document) -> Document:
        raise NotImplementedError

    @abc.abstractmethod
    def evaluate(self, act_document: Document, pred_document: Document,
                 only_content=False) -> dict:
        """ General evaluate functionality by tag, field and group level
        :param act_document: actual document (ground truth)
        :param pred_document: predicted document
        :param only_content: if True evaluation ignores locations of tags,
            considers only content
        """
        raise NotImplementedError

    def copy(self, document: Document, base_document: Document) -> Document:
        raise NotImplementedError

    @staticmethod
    def _post_response(doc, session, url, task_id):
        output_document_bson: dict = doc.to_dict()
        output_document_dict = replace_object_ids_with_string(
            output_document_bson)
        output_document_json = json.dumps(output_document_dict,
                                          ignore_nan=True)
        session.headers.update({"Content-Type": "application/json"})
        post_response: requests.Response = session.post(
            url + '/' + task_id, data=output_document_json, verify=False)
        return post_response

    def execute_based_on_match(self, task_id: str, base_doc_task_id: str,
                               token: str, url: str) -> requests.Response:
        session = requests.Session()
        session.headers = {'x-auth': token}

        get_response: requests.Response = session.get(url + '/' + task_id,
                                                      verify=False)
        get_response_dict: dict = get_response.json()
        doc_data_path: str = get_response_dict['documentRootPath']
        document_json: dict = get_response_dict['inputDocument']
        doc: Document = Document.from_dict(document_json,
                                           data_path=doc_data_path)

        base_doc_get_response: requests.Response = session.get(
            url + '/' + base_doc_task_id, verify=False)
        base_doc_get_response_dict: dict = base_doc_get_response.json()
        base_doc_data_path: str = base_doc_get_response_dict[
            'documentRootPath']
        base_document_json: dict = base_doc_get_response_dict['inputDocument']
        base_doc: Document = Document.from_dict(base_document_json,
                                                data_path=base_doc_data_path)

        doc = self.copy(document=doc, base_document=base_doc)
        return self._post_response(doc=doc, session=session, url=url,
                                   task_id=task_id)

    def execute_genie(
            self, task_id: str, token: str, url: str) -> requests.Response:
        """Alias for execute_genie_v2"""
        return self.execute_genie_v2(task_id=task_id, token=token, url=url)

    def execute_genie_v2(self, task_id: str, token: str,
                         url: str) -> requests.Response:
        """Execute genie for a given task_id"""
        session = requests.Session()
        session.headers = {'x-auth': token}
        get_response: requests.Response = session.get(url + '/' + task_id,
                                                      verify=False)
        get_response_dict: dict = get_response.json()
        doc_data_path: str = get_response_dict['documentRootPath']
        document_json: dict = get_response_dict['inputDocument']
        doc: Document = Document.from_dict(document_json,
                                           data_path=doc_data_path)
        doc = self.predict(doc)
        return self._post_response(doc=doc, session=session, url=url,
                                   task_id=task_id)

    def eval_tag_level(self, act_document: Document,
                       pred_document: Document, only_content=False) -> dict:
        """ Evaluation on tag level.
        Comparison is between all actual and predicted tags
            of each extraction field
        :returns: dict with keys - extraction field names and values - dict
            of computed metrics
        for that field"""

        result = {}
        for field_name in act_document.y.keys():
            tp_t, tn_t, fp_t, fn_t = 0, 0, 0, 0
            act_tags = []
            pred_tags = []
            for act_field in act_document.y[field_name]:
                act_tags.extend(act_field.tags)
            for pred_field in pred_document.y[field_name]:
                pred_tags.extend(pred_field.tags)
            if act_tags or pred_tags:
                tp_t = self.compute_tag_level_tp(
                    act_tags=act_tags, pred_tags=pred_tags,
                    only_content=only_content)
                fp_t = len(pred_tags) - tp_t
                fn_t = len(act_tags) - tp_t
                conf_matrix = ConfusionMatrix(TP=tp_t,
                                              TN=tn_t,
                                              FP=fp_t,
                                              FN=fn_t)
                result[field_name] = conf_matrix.compute_metrics()
        return result

    def eval_field_level(self, act_document: Document,
                         pred_document: Document, only_content=False) -> dict:
        """ Evaluation on field level.
        Comparison is between all actual and predicted
            fields of each extraction field.
        :returns: dict with keys - field names, values
            - dict of computed metrics for that
        field"""

        result = {}
        for field_name in act_document.y.keys():
            pred_fields = pred_document.y[field_name]
            act_fields = act_document.y[field_name]
            conf_matrix = self.get_field_level_conf_matrix(
                act_fields=act_fields, pred_fields=pred_fields,
                only_content=only_content)
            if (conf_matrix.TP == conf_matrix.TN == conf_matrix.FP
                    == conf_matrix.FN == 0):
                continue
            result[field_name] = conf_matrix.compute_metrics()
        return result

    def eval_group_level(self, act_document: Document,
                         pred_document: Document,
                         only_content=False) -> dict:
        """ Evaluation on group level
        Comparison is between all actual and predicted groups/entities.
        :returns: dict with keys - metrics names, values - computed scores"""

        matched_keys = []
        act_data_grouped = self.group_entities(act_document)
        pred_data_grouped = self.group_entities(pred_document)
        matched_act_keys = []
        matched_pred_keys = []
        if act_data_grouped or pred_data_grouped:
            matched_keys = self.detect_entity_matches(
                act_entities=act_data_grouped, pred_entities=pred_data_grouped)
        tp_g, tn_g, fp_g, fn_g = 0, 0, 0, 0
        for keys in matched_keys:
            tp_f = 0
            group_key_act = keys['group_key_act']
            group_key_pred = keys['group_key_pred']
            if (group_key_act not in matched_act_keys
                    and group_key_pred not in matched_pred_keys):
                group_field_names = act_data_grouped[group_key_act].keys()
                for field_name in group_field_names:
                    act_field = act_data_grouped[group_key_act][field_name]
                    pred_field = pred_data_grouped[group_key_pred][field_name]
                    if act_field.tags or pred_field.tags:
                        tp_f_per = self.compute_field_level_tp(
                            act_field=act_field, pred_field=pred_field,
                            only_content=only_content)
                        if tp_f_per == 1:
                            tp_f += 1
                        else:
                            break
                    elif act_field.value or pred_field.value:
                        if pred_field.value == act_field.value:
                            tp_f += 1
                        else:
                            break
                    else:
                        tp_f += 1
                if tp_f == len(group_field_names):
                    tp_g += 1
                    matched_act_keys.append(group_key_act)
                    matched_pred_keys.append(group_key_pred)
        fp_g = len(pred_data_grouped.keys()) - tp_g
        fn_g = len(act_data_grouped.keys()) - tp_g
        conf_matrix = ConfusionMatrix(TP=tp_g, TN=tn_g, FP=fp_g, FN=fn_g)
        result = conf_matrix.compute_metrics()
        return result

    def compute_tag_level_tp(self, act_tags: list, pred_tags: list,
                             only_content=False) -> int:
        """ Given lists of actual and predicted tags calculates tp,
        i.e. number of identical matches"""
        tp_t = 0
        if only_content:
            act_values = [act_tag.raw_value for act_tag in act_tags]
            pred_values = [pred_tag.raw_value for pred_tag in pred_tags]
            tp_t = len(list((Counter(act_values)
                             & Counter(pred_values)).elements()))
        else:
            matched_act_idxs = []
            matched_pred_idxs = []
            for act_tag_idx, act_tag in enumerate(act_tags):
                for pred_tag_idx, pred_tag in enumerate(pred_tags):
                    if (act_tag_idx not in matched_act_idxs
                            and pred_tag_idx not in matched_pred_idxs):
                        if (self.matches(act_tag=act_tag, pred_tag=pred_tag)
                                and act_tag.raw_ocr_value
                                == pred_tag.raw_ocr_value):
                            tp_t += 1
                            matched_act_idxs.append(act_tag_idx)
                            matched_pred_idxs.append(pred_tag_idx)
        return tp_t

    def get_field_level_conf_matrix(self, act_fields: list,
                                    pred_fields: list,
                                    only_content=False) -> ConfusionMatrix:
        """ Given two lists of actual and predicted fields,
            computes ConfusionMatrix.
            Takes into account both classification and extraction fields.
            Ignores empty fields"""

        tp_f, tn_f, fp_f, fn_f = 0, 0, 0, 0
        if not any(act_field.tags for act_field in act_fields) and not any(
                pred_field.tags for pred_field in pred_fields):
            act_values = [field.value for field in act_fields]
            pred_values = [field.value for field in pred_fields]
            num_empty_act_fields = len(
                [field for field in act_fields if not field.value])
            num_empty_pred_fields = len(
                [field for field in pred_fields if not field.value])
            tp_f = len(list((Counter(act_values)
                             & Counter(pred_values)).elements()))
        else:
            num_empty_act_fields = len(
                [field for field in act_fields if not field.tags])
            num_empty_pred_fields = len(
                [field for field in pred_fields if not field.tags])
            matched_act_idxs = []
            matched_pred_idxs = []
            for act_field_idx, act_field in enumerate(act_fields):
                for pred_field_idx, pred_field in enumerate(pred_fields):
                    if (act_field_idx not in matched_act_idxs
                            and pred_field_idx not in matched_pred_idxs):
                        tp_f_one = self.compute_field_level_tp(
                            act_field=act_field, pred_field=pred_field,
                            only_content=only_content)
                        tp_f += tp_f_one
                        if tp_f_one == 1:
                            matched_act_idxs.append(act_field_idx)
                            matched_pred_idxs.append(pred_field_idx)
                            break
        fp_f = len(pred_fields) - tp_f - num_empty_pred_fields
        fn_f = len(act_fields) - tp_f - num_empty_act_fields
        return ConfusionMatrix(TP=tp_f, TN=tn_f, FP=fp_f, FN=fn_f)

    def compute_field_level_tp(self,
                               act_field,
                               pred_field,
                               only_content=False) -> int:
        """ Given 2 extraction fields, actual and predicted,
        calculates tp, i.e. number of identical field matches"""
        tp_f, tp_t = 0, 0
        act_tags = [act_tag for act_tag in act_field.tags if act_tag]
        pred_tags = [pred_tag for pred_tag in pred_field.tags if pred_tag]
        if not act_tags or not pred_tags:
            return tp_f
        if len(act_tags) == len(pred_tags):
            tp_t = self.compute_tag_level_tp(act_tags=act_tags,
                                             pred_tags=pred_tags,
                                             only_content=only_content)
            if tp_t == len(act_tags):
                tp_f = 1
                return tp_f
        return tp_f

    def detect_entity_matches(self, act_entities: dict,
                              pred_entities: dict) -> list:
        """ Given actual and predicted groups/entities, finds pairs which have
        any match (even one tag in one of the fields)
        :return matched_keys: list of dicts, where each dict contains
            matched pairs' group keys """

        matched_keys = []
        for group_key_act, entity_act in act_entities.items():
            for group_key_pred, entity_pred in pred_entities.items():
                if set(entity_act.keys()) == set(entity_pred.keys()):
                    for field_name in entity_act.keys():
                        if (entity_act[field_name].tags
                                or entity_pred[field_name].tags):
                            if any(self.matches(act_tag=act_tag,
                                                pred_tag=pred_tag)
                                   for act_tag in entity_act[field_name].tags
                                   for pred_tag
                                   in entity_pred[field_name].tags):
                                matched_keys.append(
                                    {'group_key_act': group_key_act,
                                     'group_key_pred': group_key_pred})
                                break
                        elif (entity_act[field_name].value
                              == entity_pred[field_name].value):
                            matched_keys.append(
                                {'group_key_act': group_key_act,
                                 'group_key_pred': group_key_pred})
                            break
        return matched_keys

    @staticmethod
    def group_entities(document: Document) -> dict:
        """
        Group the entities by group keys.
            Fields having the same group key belong to the same entity.
            Returns dict of dicts, where keys are group keys of found entities,
            values are dicts with field names (keys): fields (values) belonging
            to that entity
        :param document:
        :return: dict of dicts
        """
        group_keys = set()
        groups = defaultdict(dict)
        for field_name in document.y.keys():
            group_keys.update(
                [item.group_key for item in document.y[field_name]
                 if item.group_key])
        for key in group_keys:
            groups[key] = {}
            for field_name in document.y.keys():
                for item in document.y[field_name]:
                    if item.group_key == key:
                        groups[key][field_name] = item
        return groups

    @staticmethod
    def matches(act_tag: ExtractionTag, pred_tag: ExtractionTag,
                threshold: float = 0.6) -> bool:
        """ Detects if there is a match between two extraction tags
            having the same page number. Returns true if
        intersection is greater than the threshold"""
        return act_tag.page.page_number == pred_tag.page.page_number and (
                act_tag & pred_tag) / min(
            act_tag, pred_tag, key=lambda x: x.area).area >= threshold

    def execute_eval(self,
                     token: str,
                     url: str,
                     model_version: str
                     ) -> List[requests.Response]:
        """ Execute evaluation for a given model_version """
        session = requests.Session()
        session.headers = {'x-auth': token}

        ground_truth_ids = session.get(
            url=f'{url}/groundtruths/{model_version}',
            verify=False
        ).json()

        post_responses = []
        for gt_id in ground_truth_ids:
            ground_truth_model_task = session.get(
                url=f'{url}/evaluations/{model_version}/{gt_id}',
                verify=False
            ).json()

            act_doc, pred_doc = self._get_doc_pair(ground_truth_model_task)
            eval_result = self.evaluate(act_document=act_doc,
                                        pred_document=pred_doc,
                                        only_content=False)

            data = {
                'modelVersion': model_version,
                'groundTruth': gt_id,
                'modelTask': ground_truth_model_task["modelTask"]["_id"],
                'evaluation': eval_result,
                'gitHash': ground_truth_model_task["modelTask"]['gitHash']
            }

            post_response = self._post_response_eval(
                session=session,
                endpoint=f'{url}/evaluations',
                data=data)
            post_responses.append(post_response)
        return post_responses

    @staticmethod
    def _get_doc_pair(data: dict) -> Tuple[Document, Document]:
        """Return the pair of actual and predicted Document objects
        from the dictionary of ground truth and model task pair."""
        pred_doc = Document.from_dict(
            raw=data["modelTask"]["outputDocument"],
            data_path=data["modelTask"]["documentRootPath"])
        act_doc = Document.from_dict(
            raw={
                "input_fields": {},
                "output_fields": data["groundTruth"]["fields"],
                "metadata": data["modelTask"]["outputDocument"]["metadata"]
            },
            data_path=data["modelTask"]["documentRootPath"])
        return act_doc, pred_doc

    @staticmethod
    def _post_response_eval(session: requests.Session,
                            endpoint: str,
                            data: dict
                            ) -> requests.Response:
        session.headers.update({"Content-Type": "application/json"})
        post_response = session.post(endpoint,
                                     data=json.dumps(data, ignore_nan=True),
                                     verify=False)
        return post_response



import itertools
from typing import Optional, Dict, List, Type

import logging


from pycognaize.common.enums import (
    IqDocumentKeysEnum,
    ID,
    IqDataTypesEnum,
    IqFieldKeyEnum
)
from pycognaize.common.utils import (
    convert_tag_coords_to_percentages,
    get_index_of_first_non_empty_list
)
from pycognaize.document.field import Field
from pycognaize.document.tag import TableTag
from pycognaize.document.tag.tag import Tag
from pycognaize.document.page import Page


class TableField(Field):
    """Base class for all pycognaize table fields"""
    tag_class: Type[Tag] = TableTag

    def __init__(self,
                 name: str,
                 tag: Optional[TableTag] = None,
                 field_id: Optional[str] = None,
                 group_key: str = None,
                 confidence: Optional[float] = -1.0
                 ):
        tags = [] if tag is None else [tag]
        super().__init__(name=name, tags=tags, group_key=group_key,
                         confidence=confidence)
        self._field_id = field_id

    def get_table_title(self, n_lines_above=8, margin=10) -> str:
        """Return the title of the table found on the pdf"""
        h = self.tags[0].page.image_height
        w = self.tags[0].page.image_width
        tags_converted = convert_tag_coords_to_percentages(
            self.tags[0], w=w, h=h)
        table_top = tags_converted['top']
        all_rows_above = []
        for line in self.tags[0].page.lines:
            all_rows_above.append(
                [w['ocr_text'] for w in line
                 if w['bottom'] < table_top + margin])
        index_of_first_non_empty_line = get_index_of_first_non_empty_list(
            all_rows_above)
        all_rows_above = all_rows_above[:index_of_first_non_empty_line + 1]
        title = ' '.join(
            itertools.chain.from_iterable(all_rows_above[-n_lines_above:]))
        return title

    @classmethod
    def construct_from_raw(
            cls, raw: dict, pages: Dict[int, Page]) -> 'TableField':
        """Create TableField object from dictionary"""
        tag_dicts: List[dict] = raw[IqDocumentKeysEnum.tags.value]
        tags = []
        for i in tag_dicts:
            try:
                tags.append(cls.tag_class.construct_from_raw(
                    raw=i, page=pages[i['page']]))
            except Exception as e:
                logging.debug(f"Failed creating tag for field {raw[ID]}: {e}")
        if len(tags) > 1:
            raise ValueError(
                f"{cls.__name__} cannot have {len(tags)}"
                f" {cls.tag_class.__name__}s")
        return cls(name=raw[IqDocumentKeysEnum.name.value],
                   tag=tags[0] if tags else None,
                   field_id=str(raw[ID]),
                   group_key=raw.get(IqFieldKeyEnum.group_key.value, '')
                   )

    def to_dict(self) -> dict:
        """Converts TableField object to dictionary"""
        field_dict = super().to_dict()
        field_dict[ID] = self._field_id
        field_dict[
            IqFieldKeyEnum.data_type.value] = IqDataTypesEnum.table.value
        field_dict[IqFieldKeyEnum.value.value] = ''
        return field_dict

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name}>"

    def __str__(self):
        return self.__repr__()



import abc
import logging
import math
from typing import Union, Tuple

from pycognaize.common.utils import convert_coord_to_num
from pycognaize.document.tag.cell import Cell

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pycognaize.document.page import Page


class Tag(metaclass=abc.ABCMeta):
    """Represents field's coordinate data on document"""

    def __init__(self,
                 left: Union[int, float],
                 right: Union[int, float],
                 top: Union[int, float],
                 bottom: Union[int, float],
                 page: 'Page'):
        """Creates and validates coordinate data"""

        self._left = left
        self._right = right
        self._top = top
        self._bottom = bottom
        self._page = page
        self.__validate()
        # Properties
        self._height = None
        self._width = None
        self._area = None
        self._xcenter = None
        self._center = None
        self._ycenter = None

    @classmethod
    @abc.abstractmethod
    def construct_from_raw(cls, raw: dict, page: 'Page') -> 'Tag':
        ...

    def __repr__(self):
        return (f"<{self.__class__.__name__}:"
                f" left: {self.left}, right: {self.right},"
                f" top: {self.top}, bottom: {self.bottom}>")

    @staticmethod
    def _parse_position(val: Union[float, int, str],
                        val_name: str) -> Union[int, float]:
        """Parse a number into a proper coordinate value

        :param val: Integer, float or a string
        :param val_name: String representing the position name
            (left/right/top/bottom)
        :return: Parsed numeric value (int or float)
        """
        try:
            return convert_coord_to_num(val)
        except TypeError as e:
            raise TypeError(f"Invalid type for {val_name} coordinate: {e}")
        except ValueError as e:
            raise ValueError(f"Failed to parse {val_name} coordinate: {e}")

    def _parse_coordinates(self):
        """If the input coordinates are in string format,
            convert them to numbers"""
        self._left = self._parse_position(val=self.left, val_name='left')
        self._right = self._parse_position(val=self.right, val_name='right')
        self._top = self._parse_position(val=self.top, val_name='top')
        self._bottom = self._parse_position(val=self.bottom, val_name='bottom')

    def __validate(self):
        """Run all input validations and necessary conversions"""
        self._validate_types()
        self._validate_coords()
        self._validate_ranges()

    def _validate_types(self):
        """Validates the input types"""
        if not isinstance(self.left, (int, float)):
            raise TypeError(f"Invalid type for left coord: {type(self.left)}")
        if not isinstance(self.right, (int, float)):
            raise TypeError(
                f"Invalid type for right coord: {type(self.right)}")
        if not isinstance(self.top, (int, float)):
            raise TypeError(f"Invalid type for top coord: {type(self.top)}")
        if not isinstance(self.bottom, (int, float)):
            raise TypeError(
                f"Invalid type for bottom coord: {type(self.bottom)}")

    def _validate_coords(self):
        """Validates the input coordinates"""
        if self.left > self.right:
            raise ValueError(f"Left ({self.left}) cannot be"
                             f" bigger than right ({self.right})")
        if self.top > self.bottom:
            raise ValueError(f"Top ({self.top}) cannot be"
                             f" bigger than bottom ({self.bottom})")
        if self.left == self.right:
            raise ValueError(f"Left ({self.left}) cannot be"
                             f" equal to right ({self.right})")
        if self.top == self.bottom:
            raise ValueError(f"Top ({self.top}) cannot be"
                             f" equal to bottom ({self.bottom})")

    def _validate_ranges(self):
        """Validates the input coordinates"""
        if not 0 <= self.left <= 100:
            logging.debug(
                f"The left coordinate should be in range 0 to 100, got"
                f" {self.left} (page {self.page.page_number})")
            self._left = min(max(0, self.left), 100)
        if not 0 <= self.right <= 100:
            logging.debug(
                f"The right coordinate should be in range 0 to 100, got"
                f" {self.right} (page {self.page.page_number})")
            self._right = min(max(0, self.right), 100)
        if not 0 <= self.top <= 100:
            logging.debug(
                f"The top coordinate should be in range 0 to 100, got"
                f" {self.top} (page {self.page.page_number})")
            self._top = min(max(0, self.top), 100)
        if not 0 <= self.bottom <= 100:
            logging.debug(
                f"The bottom coordinate should be in range 0 to 100, got"
                f" {self.bottom} (page {self.page.page_number})")
            self._bottom = min(max(0, self.bottom), 100)

    @property
    def left(self) -> Union[int, float]:
        return self._left

    @property
    def right(self) -> Union[int, float]:
        return self._right

    @property
    def top(self) -> Union[int, float]:
        return self._top

    @property
    def bottom(self) -> Union[int, float]:
        return self._bottom

    @property
    def page(self) -> 'Page':
        return self._page

    @property
    def width(self) -> Union[int, float]:
        """Width of the rectangle"""
        if self._width is None:
            self._width = self.right - self.left
        return self._width

    @property
    def height(self) -> Union[int, float]:
        """Height of the rectangle"""
        if self._height is None:
            self._height = self.bottom - self.top
        return self._height

    @property
    def area(self) -> float:
        """Area of the rectangle"""
        if self._area is None:
            self._area = self.width * self.height
        return self._area

    @property
    def xcenter(self) -> float:
        """Center of horizontal line of the rectangle"""
        if self._xcenter is None:
            self._xcenter = (self.left + self.right) / 2
        return self._xcenter

    @property
    def ycenter(self) -> float:
        """Center of vertical line of the rectangle"""
        if self._ycenter is None:
            self._ycenter = (self.top + self.bottom) / 2
        return self._ycenter

    @property
    def center(self) -> Tuple[float, float]:
        """Center point of the rectangle"""
        if self._center is None:
            self._center = (self.xcenter, self.ycenter)
        return self._center

    def intersects(self, other: Union['Tag', Cell]) -> bool:
        """Checks id there is an intersection between this and other rectangle

        :param other: Rectangle object
        """
        if isinstance(other, (Tag, Cell)):
            if (self.left < other.right and self.right > other.left
                and self.top < other.bottom and self.bottom > other.top) and \
                    self.page.page_number == other.page.page_number:
                return True
            return False
        else:
            raise NotImplementedError(
                f"Not implemented for item of type {type(other)}")

    def hshift(self, by) -> 'Tag':
        """Shifts rectangle horizontally

        :param by: the amount by which the tag should be horizontally shifted
        :return: shifted rectangle
        """
        return self.__class__(left=self.left + by, right=self.right + by,
                              top=self.top, bottom=self.bottom,
                              page=self.page)

    def vshift(self, by) -> 'Tag':
        """Shifts rectangle vertically

        :param by: the amount by which the tag should be vertically shifted
        :return: shifted rectangle
        """
        return self.__class__(left=self.left, right=self.right,
                              top=self.top + by, bottom=self.bottom + by,
                              page=self.page)

    def shift(self, horizontal, vertical) -> 'Tag':
        """Shifts rectangle by 2 axes simultaneously

        :param horizontal:
        :param vertical:
        :return: shifted rectangle
        """
        return self.hshift(horizontal).vshift(vertical)

    def __contains__(self, item: 'Tag') -> bool:
        """Checks if the item is in the rectangle

        :param item: The item to check
        """
        if isinstance(item, Tag):
            if (self.left <= item.left <= self.right and
                self.left <= item.right <= self.right and
                self.top <= item.top <= self.bottom and
                self.top <= item.bottom <= self.bottom) and \
                    self.page.page_number == item.page.page_number:
                return True
            return False
        else:
            raise NotImplementedError(
                f"Not implemented for item of type {type(item)}")

    def __and__(self, other: Union['Tag', Cell]) -> Union[int, float]:
        """The area of intersection of given rectangles"""
        if isinstance(other, (Tag, Cell)):
            if self.intersects(other):
                left = max([self.left, other.left])
                top = max([self.top, other.top])
                right = min([self.right, other.right])
                bottom = min([self.bottom, other.bottom])
                return (right - left) * (bottom - top)
            else:
                return 0
                # raise ValueError("Rectangles are disjointed")
        else:
            raise NotImplementedError(
                f"Not implemented for item of type {type(other)}")

    def __or__(self, other) -> Union[int, float]:
        """The area of union of given rectangles"""
        if isinstance(other, (Tag, Cell)):
            if self.page.page_number == other.page.page_number:
                return self.area + other.area - self.__and__(other)
            else:
                return 0
        else:
            raise NotImplementedError(
                f"Not implemented for item of type {type(other)}")

    def iou(self, other) -> Union[int, float]:
        """Calculate Intersection over Union for given rectangles"""
        if isinstance(other, (Tag, Cell)):
            if self.page.page_number == other.page.page_number:
                return (self & other) / (self | other)
            else:
                return 0
        else:
            raise NotImplementedError(
                f"Not implemented for item of type {type(other)}")

    def __add__(self, other: Union['Tag', Cell]) -> 'Tag':
        """Merge two rectangles into one"""
        if self.page.page_number == other.page.page_number:
            left = min(self.left, other.left)
            right = max(self.right, other.right)
            top = min(self.top, other.top)
            bottom = max(self.bottom, other.bottom)
            return Tag(left=left, right=right, top=top, bottom=bottom,
                       page=self.page)
        else:
            raise ValueError("Tags are not on the same page")

    def __radd__(self, other) -> 'Tag':
        """Merge two rectangles into one,
            required for using sum(<list of rectangles>)

        :param other: Another Rectangle object
        :return: Rectangle object with merged content
        """
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def is_in_rectangle(self, other: "Tag", thresh: float) -> bool:
        """Check if the rectangle is in the other rectangle

        :param other: Another rectangle object
        :param thresh: If the fraction area of the rectangle
            is in the other rectangle
            is larger or equal to thresh, it is in the other rectangle
        :return: True if rectangle is in other rectangle, False otherwise
        """
        if thresh < 0 or thresh > 1:
            raise ValueError(
                "Threshold should be a float number between 0 to 1")
        if isinstance(other, Tag):
            if self.page.page_number == other.page.page_number:
                return ((self & other) / self.area) >= thresh
            else:
                return False
        else:
            raise NotImplementedError(
                f"Not implemented for item of type {type(other)}")

    def get_top_left(self):
        return self.top, self.left

    def get_width_height(self):
        return self.width, self.height

    @abc.abstractmethod
    def to_dict(self) -> dict:
        """Return a dictionary representing the tag object"""
        pass

    def distance(self, other: 'Tag') -> Union[int, float]:
        """Return the Euclidean distance of two tag centers"""
        if self.page.page_number == other.page.page_number:
            dist = math.sqrt(
                (self.xcenter - other.xcenter)**2 +
                (self.ycenter - other.ycenter)**2)
        else:
            raise ValueError(
                "Can't compute distance between tags in different pages")
        return dist



import bson
from typing import Union

from pycognaize.common.enums import IqCellKeyEnum, ID


class Cell:
    """Represents a single element in TableTag"""

    def __init__(self, value, left_col, top_row, top, right, bottom, left,
                 col_span, row_span):
        # TODO: check for valid values
        self._value = value
        self._left_col = left_col
        self._top_row = top_row
        self._col_span = col_span
        self._row_span = row_span
        self._top = top
        self._right = right
        self._bottom = bottom
        self._left = left
        self._width = None
        self._height = None
        self._area = None

    @property
    def value(self):
        return self._value

    @property
    def left_col(self):
        return self._left_col

    @property
    def top_row(self):
        return self._top_row

    @property
    def col_span(self):
        return self._col_span

    @property
    def row_span(self):
        return self._row_span

    @property
    def top(self):
        return self._top

    @property
    def right(self):
        return self._right

    @property
    def bottom(self):
        return self._bottom

    @property
    def left(self):
        return self._left

    @property
    def area(self) -> float:
        """Area of the rectangle"""
        if self._area is None:
            self._area = self.width * self.height
        return self._area

    @property
    def width(self) -> Union[int, float]:
        """Width of the rectangle"""
        if self._width is None:
            self._width = self.right - self.left
        return self._width

    @property
    def height(self) -> Union[int, float]:
        """Height of the rectangle"""
        if self._height is None:
            self._height = self.bottom - self.top
        return self._height

    def __repr__(self):
        return (
            f'<{self.__class__.__name__}:'
            f' coords: ({self.top:<9.5f}, {self.right:<9.5f},'
            f' {self.bottom:<9.5f}, {self.left:<9.5f})'
            f' spans: ({self.row_span:<3}, {self.col_span:<3})'
            f' corner coords: ({self.left_col:<3}, {self.top_row:<3})'
            f' value: {self.value}>'
        )

    def __str__(self):
        return f'Left: {self._left_col}, Top: {self._top_row}'

    def __eq__(self, obj):
        """Adjusts method for equating with strings and other cell objects"""
        if isinstance(obj, str):
            return self._value == obj
        elif isinstance(obj, Cell):
            return (self._value == obj.value and
                    self._left_col == obj.left_col and
                    self._top_row == obj.top_row and
                    self._col_span == obj.col_span and
                    self._row_span == obj.row_span)

    def to_dict(self) -> dict:
        """Converts cell to dict"""
        return {
            ID: str(bson.ObjectId()),
            IqCellKeyEnum.left_col_top_row.value: f"{self.left_col}"
                                                  f":{self.top_row}",
            IqCellKeyEnum.col_span.value: self.col_span,
            IqCellKeyEnum.row_span.value: self.row_span,
            IqCellKeyEnum.text.value: self.value,
            IqCellKeyEnum.left.value: self.left,
            IqCellKeyEnum.top.value: self.top,
            IqCellKeyEnum.width.value: self.right - self.left,
            IqCellKeyEnum.height.value: self.bottom - self.top
        }



import os
from typing import Mapping


from pycognaize.common.enums import EnvConfigEnum
from pycognaize.common.lazy_dict import LazyDocumentDict


class Snapshot:
    """A snapshot of annotated documents from one or more collections"""

    def __init__(self, path: str):
        self._path = path
        self._documents = LazyDocumentDict(doc_path=path, data_path=path)

    @property
    def documents(self) -> Mapping:
        """Mapping of document ids to documents
        :return Mapping:  LazyDocumentDict Object
        """
        return self._documents

    @classmethod
    def _snapshot_path(cls) -> str:
        """Identify and return the snapshot path"""
        snapshot_dir = os.environ[EnvConfigEnum.SNAPSHOT_PATH.value]
        snapshot_id = os.environ[EnvConfigEnum.SNAPSHOT_ID.value]
        snapshot_path = os.path.join(snapshot_dir, snapshot_id)
        return snapshot_path

    @classmethod
    def get(cls) -> 'Snapshot':
        """Read the snapshot object from local storage and return it"""
        return cls(path=cls._snapshot_path())
